'use strict';

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

/**
 * How it works:
 * 1. It makes initialization - gets config, called template factories, sets flag to default state.
 * 2. Plugin traverse in 3 ways:
 *      - basic function declarations (FunctionDeclaration, ArrowFunctionExpression, etc.)
 *      - variable declarations with declared function
 *      - return statement with function as argument
 * 3. It tries to find closest block comment, that can be relative to this function.
 * 4. It tries to parse it with doctrine (https://github.com/eslint/doctrine).
 * 5. It makes function body normalization (removes "(a) => a * a" cases, makes it multiline)
 * 6. It inserts type check function call for arguments.
 * 7. It makes traverse for all return statements, relative to this function.
 * 8. It adds type check function call into return statement.
 * 9. If some function were transformed by plugin, on exit of Program it adds type check function declaration to file.
 */

var config = require('../config.json');
var findComment = require('./lib/find-comment');
var parseJsDoc = require('./lib/parse-jsdoc');
var normalizeFunctionBody = require('./lib/normalize-function-body');
var insertParametersCheck = require('./lib/insert-parameters-check');
var typecheckTemplate = require('./lib/typecheck-function-template');

var VISITORS = ['ClassMethod', 'ObjectMethod', 'FunctionDeclaration', 'ArrowFunctionExpression'];

module.exports = function (_ref) {
    var t = _ref.types;

    var typecheckFunctionDeclaration = typecheckTemplate.function(config.functionName);
    var typecheckFunctionCall = typecheckTemplate.call(config.functionName, t);

    var shouldInjectFunction = false;

    /**
     * @param {String} comment
     * @param {Object} path
     */
    function executeFunctionTransformation(comment, path) {
        var jsDoc = parseJsDoc(comment);

        if (!jsDoc) {
            return;
        }

        var functionName = jsDoc.name || (path.node.id ? path.node.id.name : config.defaultFunctionName);

        normalizeFunctionBody(path);
        insertParametersCheck(functionName, typecheckFunctionCall, path, jsDoc, t);

        if (jsDoc.returnStatement) {
            path.traverse(returnTypecheckVisitor, {
                jsDoc: jsDoc,
                functionPath: path,
                functionName: functionName
            });
        }

        shouldInjectFunction = true;
    }

    var returnTypecheckVisitor = {
        ReturnStatement: function ReturnStatement(path) {
            if (path.getFunctionParent().node !== this.functionPath.node) {
                return;
            }

            var argument = path.get('argument');

            argument.replaceWith(typecheckFunctionCall(this.functionName, 'return', argument.node, t.stringLiteral(JSON.stringify(this.jsDoc.returnStatement))));
        }
    };

    return {
        visitor: _defineProperty({
            Program: {
                enter: function enter() {
                    shouldInjectFunction = false;
                },
                exit: function exit(path) {
                    if (!shouldInjectFunction) {
                        return;
                    }

                    path.pushContainer('body', typecheckFunctionDeclaration());
                }
            },

            VariableDeclaration: function VariableDeclaration(path, state) {
                var comment = findComment(path, state);

                if (!comment) {
                    return;
                }

                var declarations = path.get('declarations');
                var functionDeclaration = declarations.find(function (declaration) {
                    return declaration.get('init').isFunction();
                });

                if (!functionDeclaration) {
                    return;
                }

                executeFunctionTransformation(comment, functionDeclaration.get('init'));
            },
            ReturnStatement: function ReturnStatement(path, state) {
                var argument = path.get('argument');

                if (!argument || !argument.isFunction()) {
                    return;
                }

                var comment = findComment(path, state);

                if (!comment) {
                    return;
                }

                executeFunctionTransformation(comment, argument);
            }
        }, VISITORS.join('|'), function (path, state) {
            var comment = findComment(path, state);

            if (!comment) {
                return;
            }

            executeFunctionTransformation(comment, path);
        })
    };
};